open Msh_ast

let () =
  Sys.set_signal Sys.sigint
    (Sys.Signal_handle (fun _ -> Format.eprintf "^C@."))

let expansion l =
  List.fold_right
    (fun x acc ->
      match x with
      | "~" ->
	  let home =
	    try Sys.getenv "HOME"
            with Not_found -> assert false
	  in
	  home :: acc
      | "*" ->
	  let files = Sys.readdir "."  in
	  (List.filter (fun f -> f.[0] = '.') (Array.to_list files)) @ acc
      | _ -> x :: acc)
    l []

let apply_redirect red =
  match red with
  | Input f ->
      let fd = Unix.openfile f [ Unix.O_RDONLY ] 0 in
      Unix.dup2 fd Unix.stdin;
      Unix.close fd
  | Output f ->
      let fd =
	Unix.openfile f [ Unix.O_WRONLY; Unix.O_TRUNC; Unix.O_CREAT] 0o666
      in
      Unix.dup2 fd Unix.stdout;
      Unix.close fd
  | Output_append f ->
      let fd =
	Unix.openfile f [ Unix.O_WRONLY; Unix.O_APPEND; Unix.O_CREAT] 0o666
      in
      Unix.dup2 fd Unix.stdout;
      Unix.close fd

let get_status program status =
  match status with
  | Unix.WEXITED status ->
      Format.ifprintf Format.err_formatter
	  "%s exited with code %d@." program status;
      status
  | Unix.WSIGNALED signal ->
      Format.ifprintf Format.err_formatter
	  "%s killed by signal %d@." program signal;
      signal
  | Unix.WSTOPPED signal ->
      Format.ifprintf Format.err_formatter
	  "%s stopped (???)@." program;
      signal

let rec my_wait () =
  try Unix.wait ()
  with Unix.Unix_error (Unix.EINTR, _, _) -> my_wait()

let exec_simple_cmd cmd =
  let line = expansion cmd.cmd_line in
  match line with
  | "cd" :: dir :: _ ->
      Unix.chdir dir;
      0
  | [ "cd" ] ->
      let home =
	try Sys.getenv "HOME"
        with Not_found -> assert false
      in
      Unix.chdir home;
      0
  | _ ->
      let argv = Array.of_list line in
      begin match Unix.fork () with
      | 0 ->
	  Sys.set_signal Sys.sigint Sys.Signal_default;
	  List.iter apply_redirect cmd.cmd_redirects;
	  Unix.handle_unix_error (Unix.execvp argv.(0)) argv
      | pid_son ->
          Format.ifprintf Format.err_formatter
	      "Program %n launched@." pid_son;
	  if cmd.cmd_bg then 0
          else begin
	    Sys.set_signal Sys.sigint
	      (Sys.Signal_handle (fun _ -> Unix.kill pid_son Sys.sigint));
            let rec wait_for_son () =
              let pid, status = my_wait () in
              if pid = pid_son then
                let p = "Program " ^ (string_of_int pid) in
                get_status p status
              else
                let p = "Background program " ^ (string_of_int pid) in
                ignore (get_status p status);
                wait_for_son ()
	    in
            let code = wait_for_son () in
	    Sys.set_signal Sys.sigint
	      (Sys.Signal_handle (fun _ -> Format.eprintf "^C@."));
	    code
	  end
      end



let rec eval_cmd cmd =
  match cmd with
  | Nothing -> 0
  | Simple_command scmd ->
      exec_simple_cmd scmd
  | If (cmd1, cmd2, cmd3) ->
      if eval_cmd cmd1 = 0 then
	eval_cmd cmd2
      else
	eval_cmd cmd3
  | Connection (cmd1, Seq, cmd2) ->
      ignore (eval_cmd cmd1);
      eval_cmd cmd2
  | Connection (cmd1, And, cmd2) ->
      let res1 = eval_cmd cmd1 in
      if res1 = 0 then
	eval_cmd cmd2
      else
	res1
  | Connection (cmd1, Or, cmd2) ->
      let res1 = eval_cmd cmd1 in
      if res1 = 0 then
	res1
      else
	eval_cmd cmd2
  | Connection (cmd1, Pipe, cmd2) ->
      begin match Unix.fork () with
      | 0 ->
	  let p_out, p_in = Unix.handle_unix_error Unix.pipe () in
	  begin match Unix.fork () with
	  | 0 ->
	      Unix.close p_out;
	      Unix.dup2 p_in Unix.stdout;
	      Unix.close p_in;
	      let code = eval_cmd cmd1 in
	      exit code
	  | pid_son ->
	      Unix.close p_in;
	      Unix.dup2 p_out Unix.stdin;
	      Unix.close p_out;
	      let code = eval_cmd cmd2 in
	      exit code
	  end
      | pid_son ->
          Format.ifprintf Format.err_formatter
	      "Pipe %n launched@." pid_son;
	  Sys.set_signal Sys.sigint
	    (Sys.Signal_handle (fun _ -> Unix.kill pid_son Sys.sigint));
          let rec wait_for_son () =
            let pid, status = my_wait () in
            if pid = pid_son then
              let p = "Pipe " ^ (string_of_int pid) in
              get_status p status
            else
              let p = "Background program " ^ (string_of_int pid) in
              ignore (get_status p status);
              wait_for_son ()
	  in
          let code = wait_for_son () in
	  Sys.set_signal Sys.sigint
	    (Sys.Signal_handle (fun _ -> Format.eprintf "^C@."));
	  code
      end
